
[Setting name="Show Mod Icons"]
bool Setting_ShowModIcons = true;

// not implemented yet
[Setting name="Hide deleted messages"]
bool Setting_HideDeletedMessages = false;

[Setting name="Light Chat colors"]
bool Setting_LightChatColors = false;

// not implemented yet
[Setting name="Enable whispers (needs Twitch permission)"]
bool Setting_EnableWhisper = false;

int g_pluginId;
bool g_showWindow;

class ChatMessage {
	string m_badgeText = "";
	string m_color;
	string m_username;
	string m_separator = "\\$z: ";
	string m_text;
	string m_id;
	int m_badges;
	MessageDisplayType m_type = MessageDisplayType::ChatMessage;
}

class Channel {
	array<ChatMessage@> m_messages;
	string m_name;
	int m_userFlags = 0;
	int m_roomFlags = 0;
	string m_sendingMessage;
}

enum MessageDisplayType {
	ChatMessage,
	UserNotice,
	Notice
}

enum MessageBadges {
	Admin = 1,
	Bits = 2,
	Broadcaster = 4,
	Moderator = 8,
	Subscriber = 16,
	Staff = 32,
	Turbo = 64,
	Premium = 128,
	Vip = 256,
	SubGifter = 512,
	Partner = 1024,
	HypeTrain = 2048,
	BitsLeader = 4096,
	Blue = 8192,
	Pink = 16384
}

enum ChannelFlags {
	EmoteOnly = 1,
	FollowerOnly = 2,
	R9k = 4,
	SlowMode = 8,
	SubsOnly = 16
}

class SentMessage {
	string m_text;
	string m_channel;
	uint64 m_timestamp = Time::Now;

	SentMessage(string channel,string text) {
		m_text = text;
		m_channel = channel;
	}
}

array<Channel@> g_whispers;
Channel@ g_currentChannel;
string g_newWhisper;
array<Channel@> g_mainChannels;
array<SentMessage@> g_awaitingReply;

void Main() {
	
	// 2-part channel setup
	while (!Twitch::ChannelsJoined()) yield();
	array<Twitch::ChannelState@> channels = Twitch::GetJoinedChannels();
	for (uint i = 0; i < channels.Length; i++) {
		Channel@ mainChannel = Channel();
		mainChannel.m_name = channels[i].m_name;
		mainChannel.m_messages.Reserve(256);
		g_mainChannels.InsertLast(mainChannel);
	}

	UpdateStates(channels);

	while (true) {
		if (g_pluginId == 0) g_pluginId = Twitch::Register(Twitch::MessageType::ChatMessage | Twitch::MessageType::UserState | Twitch::MessageType::UserNotice | Twitch::MessageType::Whisper);
		array<Twitch::Message@> newMessages = Twitch::Fetch(g_pluginId);
		for (uint i = 0; i < newMessages.Length; i++) {
			if (newMessages[i].m_messageType == Twitch::MessageType::ChatMessage) {
				AddChatMessageToChannels(newMessages[i]);
			} else if (newMessages[i].m_messageType == Twitch::MessageType::UserState) {
				for (uint j = 0; j < g_awaitingReply.Length; j++) {
					if (newMessages[i].m_channel == g_awaitingReply[j].m_channel) {
						newMessages[i].m_text = g_awaitingReply[j].m_text;
						g_awaitingReply.RemoveAt(j);
						AddChatMessageToChannels(newMessages[i]);
					}
				}
			} else if (newMessages[i].m_messageType == Twitch::MessageType::Whisper) {
				// Add whisper to whisper channel or add one.
				// Not implemented cause couldn't test yet
				// Maybe in a future version, who knows.
			}
		}
		yield();
	}
}

void AddChatMessageToChannels(const Twitch::Message@ &in twitchMessage) {
	ChatMessage@ message = ChatMessage();

	message.m_color = "\\$999";
	if (twitchMessage.m_tags.Exists("color") && string(twitchMessage.m_tags["color"]) != "") {
		string nameColor = string(twitchMessage.m_tags["color"]).ToLower();
		message.m_color = "\\$" + nameColor.SubStr(1,1) + nameColor.SubStr(3,1) + nameColor.SubStr(5,1);
	}
	message.m_username = twitchMessage.m_username;
	message.m_text = twitchMessage.m_text;
	if (message.m_text.Length > 8) {
		if (message.m_text[0] == 1 && message.m_text[twitchMessage.m_text.Length - 1] == 1 && message.m_text.SubStr(1,6) == "ACTION") {
			message.m_separator = " ";
			message.m_text = message.m_text.SubStr(8,message.m_text.Length - 9);
		}
	}
	
	twitchMessage.m_tags.Get("id", message.m_id);
	
	message.m_badges = GetBadgeFlags(twitchMessage);
	
	if ((message.m_badges & MessageBadges::Admin) > 0) {
		message.m_badgeText += "\\$d70" + Icons::Fire + " ";
	} else if ((message.m_badges & MessageBadges::Staff) > 0) {
		message.m_badgeText += "\\$ddd" + Icons::Wrench + " ";
	}
	if ((message.m_badges & MessageBadges::Blue) > 0) {
		message.m_badgeText += "\\$37f" + Icons::BowlingBall + " ";
	} else if ((message.m_badges & MessageBadges::Pink) > 0) {
		message.m_badgeText += "\\$f09" + Icons::BowlingBall + " ";
	}
	if ((message.m_badges & MessageBadges::Broadcaster) > 0) {
		message.m_badgeText += "\\$e11" + Icons::Video + " ";
	} else if ((message.m_badges & MessageBadges::Moderator) > 0) {
		message.m_badgeText += "\\$0a0" + Icons::Magic + " ";
	} else if ((message.m_badges & MessageBadges::Vip) > 0) {
		message.m_badgeText += "\\$e0b" + Icons::Gem + " ";
	}
	if ((message.m_badges & MessageBadges::Subscriber) > 0) {
		message.m_badgeText += "\\$a5f" + Icons::Star + " ";
	}
	if ((message.m_badges & MessageBadges::Partner) > 0) {
		message.m_badgeText += "\\$94f" + Icons::CheckCircle + " ";
	} else if ((message.m_badges & MessageBadges::Turbo) > 0) {
		message.m_badgeText += "\\$539" + Icons::BatteryFull + " ";
	} else if ((message.m_badges & MessageBadges::Premium) > 0) {
		message.m_badgeText += "\\$0ad" + Icons::Twitch + " ";
	} else if ((message.m_badges & MessageBadges::Bits) > 0) {
		message.m_badgeText += "\\$4db" + Icons::MoneyBillAlt + " ";
	} else if ((message.m_badges & MessageBadges::BitsLeader) > 0) {
		message.m_badgeText += "\\$4db" + Icons::Trophy + " ";
	} else if ((message.m_badges & MessageBadges::SubGifter) > 0) {
		message.m_badgeText += "\\$4db" + Icons::Gift + " ";
	} else if ((message.m_badges & MessageBadges::HypeTrain) > 0) {
		message.m_badgeText += "\\$eb4" + Icons::Train + " ";
	}
	
	for (uint j = 0; j < g_mainChannels.Length; j++) {
		if (twitchMessage.m_channel == g_mainChannels[j].m_name) {
			g_mainChannels[j].m_messages.InsertLast(message);
			if (g_mainChannels[j].m_messages.Length > 255) g_mainChannels[j].m_messages.RemoveAt(0);
			break;
		}
	}
}

void UpdateStates(const array<Twitch::ChannelState@> &in channels) {
	if (g_mainChannels.Length < 1) return;
	
	// Not the best performance but this will most likely only be run in the start of the plugin
	// If not, that's why we made this external lol
	for (uint i = 0; i < channels.Length; i++) {
		for (uint j = 0; j < g_mainChannels.Length; j++) {
			if (channels[i].m_name == g_mainChannels[j].m_name) {
				g_mainChannels[j].m_userFlags = GetBadgeFlags(channels[i].m_userstate);
				g_mainChannels[j].m_roomFlags = GetRoomFlags(channels[i].m_roomstate, g_mainChannels[j].m_roomFlags);
				break;
			}
		}
	}
}

int GetBadgeFlags(const Twitch::Message@ &in message) {
	int userFlags = 0;

	if (message.m_messageType == Twitch::MessageType::UserState || message.m_messageType == Twitch::MessageType::ChatMessage) {
		if (message.m_tags.Exists("badges")) {
			string badges = string(message.m_tags["badges"]);
			if (badges.Contains("admin/")) userFlags |= MessageBadges::Admin;
			if (badges.Contains("bits/")) userFlags |= MessageBadges::Bits;
			if (badges.Contains("broadcaster/")) userFlags |= MessageBadges::Broadcaster;
			if (badges.Contains("moderator/") || badges.Contains("global_mod/")) userFlags |= MessageBadges::Moderator;
			if (badges.Contains("subscriber/") || badges.Contains("founder/")) userFlags |= MessageBadges::Subscriber;
			if (badges.Contains("staff/")) userFlags |= MessageBadges::Staff;
			if (badges.Contains("turbo/")) userFlags |= MessageBadges::Turbo;
			if (badges.Contains("premium/")) userFlags |= MessageBadges::Premium;
			if (badges.Contains("vip/")) userFlags |= MessageBadges::Vip;
			if (badges.Contains("sub-gifter/")) userFlags |= MessageBadges::SubGifter;
			if (badges.Contains("partner/")) userFlags |= MessageBadges::Partner;
			if (badges.Contains("hype-train/")) userFlags |= MessageBadges::HypeTrain;
			if (badges.Contains("bits-leader/")) userFlags |= MessageBadges::BitsLeader;
			if (badges.Contains("blue-1")) userFlags |= MessageBadges::Blue;
			if (badges.Contains("pink-2")) userFlags |= MessageBadges::Pink;
		}
	}
	
	return userFlags;
}

int GetRoomFlags(const Twitch::Message@ &in message, int roomFlags) {

	if (message.m_messageType == Twitch::MessageType::RoomState) {

		// can't find a better way to do this rn? unless I can set single bits
		roomFlags = ChangeFlag(message, "emote-only", roomFlags, ChannelFlags::EmoteOnly);
		roomFlags = ChangeFlag(message, "follower-only", roomFlags, ChannelFlags::FollowerOnly);
		roomFlags = ChangeFlag(message, "r9k", roomFlags, ChannelFlags::R9k);
		roomFlags = ChangeFlag(message, "slow", roomFlags, ChannelFlags::SlowMode);
		roomFlags = ChangeFlag(message, "subs-only", roomFlags, ChannelFlags::SubsOnly);
	}

	return roomFlags;
}

int ChangeFlag(const Twitch::Message@ &in message, const string &in propertyName, int roomFlags, const int flagValue) {
	int property;

	if (message.m_tags.Get(propertyName, property)) {
		if (property == 0 && (roomFlags & flagValue) > 0) {
			roomFlags &= ~(flagValue);
		} else if (property > 0 && (roomFlags & flagValue) == 0) {
			roomFlags |= flagValue;
		}
	}
	return roomFlags;
}

void OnDisabled() {
	if (g_pluginId > 0) Twitch::Unregister(g_pluginId);
	g_pluginId = 0;
}

void RenderMenu() {
	if(UI::MenuItem(Icons::AlignLeft + " Twitch Chat Window", "", g_showWindow)) {
		g_showWindow = !g_showWindow;
	}
}

void RenderInterface() {
	if (!g_showWindow) return;

	if (UI::Begin(Icons::Brands::Twitch + " Twitch Chat###Twitch Chat", g_showWindow, UI::WindowFlags::NoCollapse | UI::WindowFlags::MenuBar)) {
		if (UI::BeginMenuBar()) {
			if (UI::BeginMenu("Settings")) {
				if (UI::MenuItem("Show mod Icons", "", Setting_ShowModIcons)) {
					Setting_ShowModIcons = !Setting_ShowModIcons;
				}
				/* @TODO future stuff
				if (UI::MenuItem("Hide deleted messages", "", Setting_HideDeletedMessages)) {
					Setting_HideDeletedMessages = !Setting_HideDeletedMessages;
				}
				*/
				UI::EndMenu();
			}
			UI::EndMenuBar();
		}
		UI::BeginTabBar("Chats");
		for (uint i = 0; i < g_mainChannels.Length; i++) {
			if (UI::BeginTabItem(Icons::Comments + " " + g_mainChannels[i].m_name)) {
				UI::BeginChild(g_mainChannels[i].m_name, vec2(0.0, -40.0));
				UI::ListClipper clipper(g_mainChannels[i].m_messages.Length);
				while (clipper.Step()) {
					for (int j = clipper.DisplayStart; j < clipper.DisplayEnd; j++) {
						if (g_mainChannels[i].m_userFlags & (MessageBadges::Moderator | MessageBadges::Broadcaster | MessageBadges::Admin) > 0 && !(g_mainChannels[i].m_messages[j].m_badges & (MessageBadges::Moderator | MessageBadges::Broadcaster | MessageBadges::Admin) > 0) && Setting_ShowModIcons) {
							if (UI::Selectable("\\$ccc" + Icons::TimesCircle, false)) {
								print("/ban " + g_mainChannels[i].m_messages[j].m_username);
							}
							UI::SameLine();
							if (UI::Selectable("\\$ccc" + Icons::Clock, false)) {
								print("/timeout " + g_mainChannels[i].m_messages[j].m_username);
							}
							UI::SameLine();
							if (UI::Selectable("\\$ccc" + Icons::TrashAlt, false)) {
								print("/delete " + g_mainChannels[i].m_messages[j].m_id);
							}
							UI::SameLine();
						}
						UI::Text(g_mainChannels[i].m_messages[j].m_badgeText + g_mainChannels[i].m_messages[j].m_color + g_mainChannels[i].m_messages[j].m_username + g_mainChannels[i].m_messages[j].m_separator + g_mainChannels[i].m_messages[j].m_text);
					}
				}
				if (UI::GetScrollY() >= UI::GetScrollMaxY()) UI::SetScrollHereY(1.0f);
				UI::EndChild();
				UI::Separator();
				bool changed = false;
				g_mainChannels[i].m_sendingMessage = UI::InputText("Chat", g_mainChannels[i].m_sendingMessage, changed, UI::InputTextFlags::EnterReturnsTrue);
				if (changed) {
					Twitch::MessageCode sentCode = Twitch::SendMessage(g_mainChannels[i].m_name, g_mainChannels[i].m_sendingMessage);
					g_awaitingReply.InsertLast(SentMessage(g_mainChannels[i].m_name, g_mainChannels[i].m_sendingMessage));
					UI::SetKeyboardFocusHere(-1);
					g_mainChannels[i].m_sendingMessage = "";
				}
				UI::EndTabItem();
			}
		}
		if (Setting_EnableWhisper) {
			for (uint i = 0; i < g_whispers.Length; i++) {
				if (UI::BeginTabItem(Icons::Envelope + " " + g_whispers[i].m_name)) {
					UI::BeginChild(g_whispers[i].m_name, vec2(0.0, -40.0));
					for (uint j = 0; j < g_whispers[i].m_messages.Length; j++) {
						UI::Text(g_whispers[i].m_messages[j].m_badgeText + g_whispers[i].m_messages[j].m_color + g_whispers[i].m_messages[j].m_username + g_whispers[i].m_messages[j].m_separator + g_whispers[i].m_messages[j].m_text);
					}
					UI::EndChild();
					UI::Separator();
					bool changed = false;
					g_whispers[i].m_sendingMessage = UI::InputText("Whisper", g_whispers[i].m_sendingMessage, changed, UI::InputTextFlags::EnterReturnsTrue);
					if (changed) {
						Twitch::MessageCode sentCode = Twitch::SendWhisper(g_whispers[i].m_name, g_whispers[i].m_sendingMessage);
						UI::SetKeyboardFocusHere(-1);
						g_whispers[i].m_sendingMessage = "";
					}
					UI::EndTabItem();
				}
			}
			if (UI::BeginTabItem("+")) {
				UI::BeginChild("New Whisper");
				bool changed = false;
				g_newWhisper = UI::InputText("Username to whisper", g_newWhisper, changed, UI::InputTextFlags::EnterReturnsTrue);
				if (changed) {
					Channel@ whisper = Channel();
					whisper.m_name = g_newWhisper;
					g_whispers.InsertLast(whisper);
					UI::SetKeyboardFocusHere(-1);
					g_newWhisper = "";
				}
				UI::EndChild();
				UI::EndTabItem();
			}
		}
		UI::EndTabBar();

		UI::End();
	}
}
